@@ -22,6 +22,10 @@
  * Most of these handlers will be auto-connected by libglade.  GtkAction
  * callbacks belong in gtkhtml-editor-actions.c. */
 
+#include "gtkhtml-editor-signals.h"
+
+#include <libgnome/gnome-url.h>
+
 #include "gtkhtml-editor-private.h"
 
 /* Think of this as a Python decorator.  It silences compiler
@@ -111,9 +115,11 @@ cell_properties_set_table (GtkhtmlEditor *editor,
 {
 	GtkHTML *html;
 	HTMLObject *parent;
+	HTMLTableCell *cell;
 	HTMLTableCell *iter;
 
 	html = gtkhtml_editor_get_html (editor);
+	cell = HTML_TABLE_CELL (editor->priv->cell_object);
 	parent = editor->priv->cell_parent;
 
 	iter = html_engine_get_table_cell (html->engine);
@@ -177,19 +183,14 @@ cell_properties_set_background_color_cb (GtkhtmlEditor *editor,
                                          HTMLTableCell *cell,
                                          GtkWidget *widget)
 {
-	GtkhtmlColorCombo *combo;
 	GtkHTML *html;
 	GdkColor color;
-	gboolean got_color;
 
-	combo = GTKHTML_COLOR_COMBO (widget);
 	html = gtkhtml_editor_get_html (editor);
+	gtkhtml_color_combo_get_current_color (
+		GTKHTML_COLOR_COMBO (widget), &color);
 
-	/* The default table cell color is transparent. */
-	got_color = gtkhtml_color_combo_get_current_color (combo, &color);
-
-	html_engine_table_cell_set_bg_color (
-		html->engine, cell, got_color ? &color : NULL);
+	html_engine_table_cell_set_bg_color (html->engine, cell, &color);
 }
 
 static void
@@ -251,7 +252,7 @@ cell_properties_set_horizontal_alignment_cb (GtkhtmlEditor *editor,
 	active = gtk_combo_box_get_active (GTK_COMBO_BOX (widget));
 	align = HTML_HALIGN_LEFT + (HTMLHAlignType) active;
 
-	html_engine_table_cell_set_halign (html->engine, cell, align);
+	html_engine_table_cell_set_halign (html->engine, cell, active);
 }
 
 static void
@@ -293,7 +294,7 @@ cell_properties_set_width_cb (GtkhtmlEditor *editor,
                               GtkWidget *widget)
 {
 	GtkHTML *html;
-	gboolean sensitive;
+	gboolean sensitive;;
 	gint active;
 	gint value;
 
@@ -462,9 +463,9 @@ gtkhtml_editor_cell_properties_show_window_cb (GtkWidget *window))
 	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
 
 	widget = WIDGET (CELL_PROPERTIES_COLOR_COMBO);
-	gtkhtml_color_combo_set_current_color (
-		GTKHTML_COLOR_COMBO (widget),
-		cell->have_bg ? &cell->bg : NULL);
+	if (cell->have_bg)
+		gtkhtml_color_combo_set_current_color (
+			GTKHTML_COLOR_COMBO (widget), &cell->bg);
 
 	widget = WIDGET (CELL_PROPERTIES_IMAGE_FILE_CHOOSER);
 	if (cell->have_bgPixmap) {
@@ -565,7 +566,6 @@ gtkhtml_editor_cell_properties_width_changed_cb (GtkWidget *window))
 	GtkAdjustment *adjustment;
 	GtkWidget *widget;
 	gboolean sensitive;
-	gdouble value;
 	gint active;
 
 	editor = extract_gtkhtml_editor (window);
@@ -582,16 +582,15 @@ gtkhtml_editor_cell_properties_width_changed_cb (GtkWidget *window))
 	gtk_widget_set_sensitive (widget, sensitive);
 
 	if (active == SIZE_UNIT_PERCENT) {
-		gtk_adjustment_set_upper (adjustment, 100.0);
+		adjustment->upper = 100;
 		gtk_adjustment_changed (adjustment);
 	} else {
-		gtk_adjustment_set_upper (adjustment, (gdouble) G_MAXINT);
+		adjustment->upper = (gdouble) G_MAXINT;
 		gtk_adjustment_changed (adjustment);
 	}
 
 	/* Clamp the value between the new bounds. */
-	value = gtk_adjustment_get_value (adjustment);
-	gtk_adjustment_set_value (adjustment, value);
+	gtk_adjustment_set_value (adjustment, adjustment->value);
 
 	cell_properties_set (
 		editor, cell_properties_set_width_cb,
@@ -782,8 +781,12 @@ gtkhtml_editor_image_properties_show_window_cb (GtkWidget *window))
 		|| html_object_get_data (parent, "template_image") == NULL)
 		&& image->image_ptr->url != NULL) {
 
-		gtk_file_chooser_set_uri (
-			GTK_FILE_CHOOSER (widget), image->image_ptr->url);
+		gchar *filename;
+
+		filename = gtk_html_filename_from_uri (image->image_ptr->url);
+		gtk_file_chooser_set_filename (
+			GTK_FILE_CHOOSER (widget), filename);
+		g_free (filename);
 	} else
 		gtk_file_chooser_unselect_all (GTK_FILE_CHOOSER (widget));
 
@@ -962,7 +965,7 @@ gtkhtml_editor_image_properties_url_button_clicked_cb (GtkWidget *window))
 	text = gtk_entry_get_text (GTK_ENTRY (widget));
 
 	g_return_if_fail (text != NULL);
-	gtkhtml_editor_show_uri (GTK_WINDOW (window), text);
+	gnome_url_show (text, NULL);
 
 	g_object_unref (editor);
 }
@@ -1012,23 +1015,6 @@ gtkhtml_editor_image_properties_url_entry_changed_cb (GtkWidget *window))
  * Link Properties Window
  *****************************************************************************/
 
-static gchar *
-sanitize_description_text (const gchar *ptext)
-{
-	gchar *p, *text = g_strdup (ptext);
-
-	if (!text)
-		return text;
-
-	/* replace line breakers with spaces */
-	for (p = text; *p; p++) {
-		if (*p == '\r' || *p == '\n')
-			*p = ' ';
-	}
-
-	return g_strstrip (text);
-}
-
 AUTOCONNECTED_SIGNAL_HANDLER (void
 gtkhtml_editor_link_properties_description_changed_cb (GtkWidget *window))
 {
@@ -1044,31 +1030,37 @@ gtkhtml_editor_link_properties_description_changed_cb (GtkWidget *window))
 	dsc_entry = WIDGET (LINK_PROPERTIES_DESCRIPTION_ENTRY);
 	url_entry = WIDGET (LINK_PROPERTIES_URL_ENTRY);
 
-	text = sanitize_description_text (gtk_entry_get_text (GTK_ENTRY (dsc_entry)));
+	text = g_strstrip (g_strdup (
+		gtk_entry_get_text (GTK_ENTRY (dsc_entry))));
 
 	editor->priv->link_custom_description = (*text != '\0');
 
 	if (editor->priv->link_custom_description) {
+		gint start_offset;
+		gint end_offset;
 		glong length;
-		Link *link;
 
-		link = html_text_get_link_at_offset (HTML_TEXT (html->engine->cursor->object), html->engine->cursor->offset);
+		start_offset = editor->priv->link_start_offset;
+		end_offset = editor->priv->link_end_offset;
 		length = g_utf8_strlen (text, -1);
 
-		if (link && link->start_offset != link->end_offset) {
+		if (start_offset != end_offset) {
 			html_cursor_jump_to (
 				html->engine->cursor, html->engine,
-				html->engine->cursor->object, link->start_offset);
+				editor->priv->link_object, start_offset);
 			html_engine_set_mark (html->engine);
 			html_cursor_jump_to (
 				html->engine->cursor, html->engine,
-				html->engine->cursor->object, link->end_offset);
+				editor->priv->link_object, end_offset);
 			html_engine_delete (html->engine);
 		}
 
 		html_engine_paste_link (
 			html->engine, text, length,
 			gtk_entry_get_text (GTK_ENTRY (url_entry)));
+
+		editor->priv->link_object = html->engine->cursor->object;
+		editor->priv->link_end_offset = start_offset + length;
 	}
 
 	g_free (text);
@@ -1096,34 +1088,11 @@ gtkhtml_editor_link_properties_url_changed_cb (GtkWidget *window))
 
 	gtk_action_set_sensitive (ACTION (TEST_URL), *text != '\0');
 
-	if (html_engine_is_selection_active (html->engine)) {
+	if (html_engine_is_selection_active (html->engine))
 		html_engine_set_link (html->engine, text);
-	} else if (!editor->priv->link_custom_description) {
-		gchar *descr = sanitize_description_text (text);
-		gtk_entry_set_text (GTK_ENTRY (dsc_entry), descr);
-		g_free (descr);
+	else if (!editor->priv->link_custom_description) {
+		gtk_entry_set_text (GTK_ENTRY (dsc_entry), text);
 		editor->priv->link_custom_description = FALSE;
-	} else {
-		glong length;
-		Link *link;
-		const gchar *descr = gtk_entry_get_text (GTK_ENTRY (dsc_entry));
-
-		link = html_text_get_link_at_offset (HTML_TEXT (html->engine->cursor->object), html->engine->cursor->offset);
-		length = g_utf8_strlen (descr, -1);
-
-		if (link && link->start_offset != link->end_offset) {
-			html_cursor_jump_to (
-				html->engine->cursor, html->engine,
-				html->engine->cursor->object, link->start_offset);
-			html_engine_set_mark (html->engine);
-			html_cursor_jump_to (
-				html->engine->cursor, html->engine,
-				html->engine->cursor->object, link->end_offset);
-			html_engine_delete (html->engine);
-		}
-
-		html_engine_paste_link (
-			html->engine, descr, length, text);
 	}
 
 	g_free (text);
@@ -1138,11 +1107,8 @@ gtkhtml_editor_link_properties_show_window_cb (GtkWidget *window))
 	GtkWidget *dsc_entry;
 	GtkWidget *url_entry;
 	GtkHTML *html;
-	gchar *url = NULL, *dsc = NULL;
+	gchar *url = NULL;
 	gboolean sensitive;
-	HTMLCursor *cursor;
-	gint start_offset = 0;
-	gint end_offset = 0;
 
 	editor = extract_gtkhtml_editor (window);
 	html = gtkhtml_editor_get_html (editor);
@@ -1150,61 +1116,57 @@ gtkhtml_editor_link_properties_show_window_cb (GtkWidget *window))
 	dsc_entry = WIDGET (LINK_PROPERTIES_DESCRIPTION_ENTRY);
 	url_entry = WIDGET (LINK_PROPERTIES_URL_ENTRY);
 
-	editor->priv->link_custom_description = FALSE;
-
-	cursor = html->engine->cursor;
-
-	if (HTML_IS_TEXT (cursor->object))
-		url = html_object_get_complete_url (
-			cursor->object, cursor->offset);
-
-	if (url != NULL) {
-		if (HTML_IS_IMAGE (cursor->object)) {
-			start_offset = 0;
-			end_offset = 1;
-		} else {
-			Link *link;
-
-			link = html_text_get_link_at_offset (
-				HTML_TEXT (cursor->object),
-				cursor->offset);
-			if (link != NULL) {
-				start_offset = link->start_offset;
-				end_offset = link->end_offset;
-				dsc = html_text_get_link_text (HTML_TEXT (cursor->object), cursor->offset);
-				editor->priv->link_custom_description = dsc && !g_str_equal (dsc, url);
-			}
-		}
-	} else if (HTML_IS_TEXT (cursor->object)) {
-		start_offset = cursor->offset;
-		end_offset = start_offset;
-	}
-
-	sensitive = (url == NULL);
-
 	if (html_engine_is_selection_active (html->engine)) {
-		if (sensitive && !dsc)
-			dsc = html_engine_get_selection_string (html->engine);
 		sensitive = FALSE;
+		editor->priv->link_object = NULL;
+	} else {
+		HTMLCursor *cursor;
+		gint start_offset = 0;
+		gint end_offset = 0;
+
+		cursor = html->engine->cursor;
+		editor->priv->link_object = cursor->object;
+
+		if (HTML_IS_TEXT (cursor->object))
+			url = html_object_get_complete_url (
+				cursor->object, cursor->offset);
+
+		if (url != NULL) {
+			if (HTML_IS_IMAGE (cursor->object)) {
+				start_offset = 0;
+				end_offset = 1;
+			} else {
+				Link *link;
+
+				link = html_text_get_link_at_offset (
+					HTML_TEXT (cursor->object),
+					cursor->offset);
+				if (link != NULL) {
+					start_offset = link->start_offset;
+					end_offset = link->end_offset;
+				}
+			}
+		} else if (HTML_IS_TEXT (cursor->object)) {
+			start_offset = cursor->offset;
+			end_offset = start_offset;
+		}
+
+		sensitive = (url == NULL);
+
+		gtk_entry_set_text (
+			GTK_ENTRY (url_entry),
+			(url != NULL) ? url : "http://");
+
+		editor->priv->link_object = cursor->object;
+		editor->priv->link_start_offset = start_offset;
+		editor->priv->link_end_offset = end_offset;
 	}
 
+	gtk_entry_set_text (GTK_ENTRY (dsc_entry), "");
 	gtk_widget_set_sensitive (dsc_entry, sensitive);
-	gtk_entry_set_text (
-		GTK_ENTRY (url_entry),
-		(url != NULL) ? url : "http://");
-
-	if (dsc) {
-		gchar *tmp = dsc;
-
-		dsc = sanitize_description_text (dsc);
-		g_free (tmp);
-	}
-
-	gtk_entry_set_text (GTK_ENTRY (dsc_entry), dsc ? dsc : "");
 	gtk_widget_grab_focus (url_entry);
 
 	g_free (url);
-	g_free (dsc);
 
 	g_object_unref (editor);
 }
@@ -1631,7 +1593,6 @@ gtkhtml_editor_rule_properties_width_changed_cb (GtkWidget *window))
 	GtkWidget *widget;
 	GtkHTML *html;
 	HTMLRule *rule;
-	gdouble value;
 	gint active;
 
 	editor = extract_gtkhtml_editor (window);
@@ -1645,22 +1606,20 @@ gtkhtml_editor_rule_properties_width_changed_cb (GtkWidget *window))
 	adjustment = gtk_spin_button_get_adjustment (GTK_SPIN_BUTTON (widget));
 
 	if (active == SIZE_UNIT_PERCENT) {
-		gtk_adjustment_set_upper (adjustment, 100.0);
+		adjustment->upper = 100;
 		gtk_adjustment_changed (adjustment);
 	} else {
-		gtk_adjustment_set_upper (adjustment, (gdouble) G_MAXINT);
+		adjustment->upper = (gdouble) G_MAXINT;
 		gtk_adjustment_changed (adjustment);
 	}
 
 	/* Clamp the value between the new bounds. */
-	value = gtk_adjustment_get_value (adjustment);
-	gtk_adjustment_set_value (adjustment, value);
-	value = gtk_adjustment_get_value (adjustment);
+	gtk_adjustment_set_value (adjustment, adjustment->value);
 
 	html_rule_set_length (
 		rule, html->engine,
-		(active == SIZE_UNIT_PX) ? (gint) value : 0,
-		(active == SIZE_UNIT_PX) ? 0 : (gint) value);
+		(active == SIZE_UNIT_PX) ? (gint) adjustment->value : 0,
+		(active == SIZE_UNIT_PX) ? 0 : (gint) adjustment->value);
 
 	g_object_unref (editor);
 }
@@ -1728,27 +1687,6 @@ gtkhtml_editor_table_properties_border_changed_cb (GtkWidget *window,
 	g_object_unref (editor);
 }
 
-AUTOCONNECTED_SIGNAL_HANDLER (void
-gtkhtml_editor_table_properties_color_changed_cb (GtkWidget *window,
-                                                  GtkhtmlColorCombo *combo))
-{
-	GtkhtmlEditor *editor;
-	HTMLTable *table;
-	GtkHTML *html;
-	GdkColor color;
-	gboolean got_color;
-
-	editor = extract_gtkhtml_editor (window);
-	html = gtkhtml_editor_get_html (editor);
-	table = HTML_TABLE (editor->priv->table_object);
-
-	/* The default table color is transparent. */
-	got_color = gtkhtml_color_combo_get_current_color (combo, &color);
-
-	html_engine_table_set_bg_color (
-		html->engine, table, got_color ? &color : NULL);
-}
-
 AUTOCONNECTED_SIGNAL_HANDLER (void
 gtkhtml_editor_table_properties_cols_changed_cb (GtkWidget *window,
                                                  GtkSpinButton *button))
@@ -1923,9 +1861,7 @@ gtkhtml_editor_table_properties_show_window_cb (GtkWidget *window))
 	widget = WIDGET (TABLE_PROPERTIES_BORDER_SPIN_BUTTON);
 	gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), value);
 
-	widget = WIDGET (TABLE_PROPERTIES_COLOR_COMBO);
-	gtkhtml_color_combo_set_current_color (
-		GTKHTML_COLOR_COMBO (widget), table->bgColor);
+	/* TODO - Initialize color combo */
 
 	if (table->bgPixmap != NULL) {
 		gchar *filename;
@@ -1957,7 +1893,6 @@ gtkhtml_editor_table_properties_width_changed_cb (GtkWidget *window))
 	GtkHTML *html;
 	HTMLTable *table;
 	gboolean sensitive;
-	gdouble value;
 	gint active;
 
 	editor = extract_gtkhtml_editor (window);
@@ -1976,21 +1911,19 @@ gtkhtml_editor_table_properties_width_changed_cb (GtkWidget *window))
 	gtk_widget_set_sensitive (widget, sensitive);
 
 	if (active == SIZE_UNIT_PERCENT) {
-		gtk_adjustment_set_upper (adjustment, 100.0);
+		adjustment->upper = 100;
 		gtk_adjustment_changed (adjustment);
 	} else {
-		gtk_adjustment_set_upper (adjustment, (gdouble) G_MAXINT);
+		adjustment->upper = (gdouble) G_MAXINT;
 		gtk_adjustment_changed (adjustment);
 	}
 
 	/* Clamp the value between the new bounds. */
-	value = gtk_adjustment_get_value (adjustment);
-	gtk_adjustment_set_value (adjustment, value);
-	value = gtk_adjustment_get_value (adjustment);
+	gtk_adjustment_set_value (adjustment, adjustment->value);
 
 	html_engine_table_set_width (
 		html->engine, table,
-		sensitive ? (gint) value : 0,
+		sensitive ? (gint) adjustment->value : 0,
 		(active == SIZE_UNIT_PERCENT));
 
 	g_object_unref (editor);
